@@ -1,10 +1,11 @@
 use crate::{
-    compilation::{compile_validators, context::CompilationContext},
+    compiler,
     error::{no_error, ErrorIterator},
     keywords::CompilationResult,
-    paths::InstancePath,
-    schema_node::SchemaNode,
-    validator::{format_iter_of_validators, format_validators, PartialApplication, Validate},
+    node::SchemaNode,
+    paths::LazyLocation,
+    validator::{PartialApplication, Validate},
+    ValidationError,
 };
 use serde_json::{Map, Value};
 
@@ -14,20 +15,35 @@ pub(crate) struct ItemsArrayValidator {
 impl ItemsArrayValidator {
     #[inline]
     pub(crate) fn compile<'a>(
+        ctx: &compiler::Context,
         schemas: &'a [Value],
-        context: &CompilationContext,
     ) -> CompilationResult<'a> {
-        let keyword_context = context.with_path("items");
+        let kctx = ctx.new_at_location("items");
         let mut items = Vec::with_capacity(schemas.len());
         for (idx, item) in schemas.iter().enumerate() {
-            let item_context = keyword_context.with_path(idx);
-            let validators = compile_validators(item, &item_context)?;
-            items.push(validators)
+            let ictx = kctx.new_at_location(idx);
+            let validators = compiler::compile(&ictx, ictx.as_resource_ref(item))?;
+            items.push(validators);
         }
         Ok(Box::new(ItemsArrayValidator { items }))
     }
 }
 impl Validate for ItemsArrayValidator {
+    #[allow(clippy::needless_collect)]
+    fn iter_errors<'i>(&self, instance: &'i Value, location: &LazyLocation) -> ErrorIterator<'i> {
+        if let Value::Array(items) = instance {
+            let errors: Vec<_> = items
+                .iter()
+                .zip(self.items.iter())
+                .enumerate()
+                .flat_map(move |(idx, (item, node))| node.iter_errors(item, &location.push(idx)))
+                .collect();
+            Box::new(errors.into_iter())
+        } else {
+            no_error()
+        }
+    }
+
     fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Array(items) = instance {
             items
@@ -39,33 +55,18 @@ impl Validate for ItemsArrayValidator {
         }
     }
 
-    #[allow(clippy::needless_collect)]
-    fn validate<'instance>(
+    fn validate<'i>(
         &self,
-        instance: &'instance Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'instance> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
         if let Value::Array(items) = instance {
-            let errors: Vec<_> = items
-                .iter()
-                .zip(self.items.iter())
-                .enumerate()
-                .flat_map(move |(idx, (item, node))| node.validate(item, &instance_path.push(idx)))
-                .collect();
-            Box::new(errors.into_iter())
-        } else {
-            no_error()
+            for (idx, (item, node)) in items.iter().zip(self.items.iter()).enumerate() {
+                let item_location = location.push(idx);
+                node.validate(item, &item_location)?;
+            }
         }
-    }
-}
-
-impl core::fmt::Display for ItemsArrayValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(
-            f,
-            "items: [{}]",
-            format_iter_of_validators(self.items.iter().map(SchemaNode::validators))
-        )
+        Ok(())
     }
 }
 
@@ -75,16 +76,27 @@ pub(crate) struct ItemsObjectValidator {
 
 impl ItemsObjectValidator {
     #[inline]
-    pub(crate) fn compile<'a>(
-        schema: &'a Value,
-        context: &CompilationContext,
-    ) -> CompilationResult<'a> {
-        let keyword_context = context.with_path("items");
-        let node = compile_validators(schema, &keyword_context)?;
+    pub(crate) fn compile<'a>(ctx: &compiler::Context, schema: &'a Value) -> CompilationResult<'a> {
+        let ctx = ctx.new_at_location("items");
+        let node = compiler::compile(&ctx, ctx.as_resource_ref(schema))?;
         Ok(Box::new(ItemsObjectValidator { node }))
     }
 }
 impl Validate for ItemsObjectValidator {
+    #[allow(clippy::needless_collect)]
+    fn iter_errors<'i>(&self, instance: &'i Value, location: &LazyLocation) -> ErrorIterator<'i> {
+        if let Value::Array(items) = instance {
+            let errors: Vec<_> = items
+                .iter()
+                .enumerate()
+                .flat_map(move |(idx, item)| self.node.iter_errors(item, &location.push(idx)))
+                .collect();
+            Box::new(errors.into_iter())
+        } else {
+            no_error()
+        }
+    }
+
     fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Array(items) = instance {
             items.iter().all(|i| self.node.is_valid(i))
@@ -93,33 +105,25 @@ impl Validate for ItemsObjectValidator {
         }
     }
 
-    #[allow(clippy::needless_collect)]
-    fn validate<'instance>(
+    fn validate<'i>(
         &self,
-        instance: &'instance Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'instance> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
         if let Value::Array(items) = instance {
-            let errors: Vec<_> = items
-                .iter()
-                .enumerate()
-                .flat_map(move |(idx, item)| self.node.validate(item, &instance_path.push(idx)))
-                .collect();
-            Box::new(errors.into_iter())
-        } else {
-            no_error()
+            for (idx, item) in items.iter().enumerate() {
+                let item_location = location.push(idx);
+                self.node.validate(item, &item_location)?;
+            }
         }
+        Ok(())
     }
 
-    fn apply<'a>(
-        &'a self,
-        instance: &Value,
-        instance_path: &InstancePath,
-    ) -> PartialApplication<'a> {
+    fn apply(&self, instance: &Value, location: &LazyLocation) -> PartialApplication {
         if let Value::Array(items) = instance {
             let mut results = Vec::with_capacity(items.len());
             for (idx, item) in items.iter().enumerate() {
-                let path = instance_path.push(idx);
+                let path = location.push(idx);
                 results.push(self.node.apply_rooted(item, &path));
             }
             let mut output: PartialApplication = results.into_iter().collect();
@@ -129,7 +133,7 @@ impl Validate for ItemsObjectValidator {
             // `ItemsObjectValidator` is not used when prefixItems is defined, this is true if
             // there are any items in the instance.
             let schema_was_applied = !items.is_empty();
-            output.annotate(serde_json::json! {schema_was_applied}.into());
+            output.annotate(serde_json::json!(schema_was_applied).into());
             output
         } else {
             PartialApplication::valid_empty()
@@ -137,12 +141,6 @@ impl Validate for ItemsObjectValidator {
     }
 }
 
-impl core::fmt::Display for ItemsObjectValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "items: {}", format_validators(self.node.validators()))
-    }
-}
-
 pub(crate) struct ItemsObjectSkipPrefixValidator {
     node: SchemaNode,
     skip_prefix: usize,
@@ -153,10 +151,10 @@ impl ItemsObjectSkipPrefixValidator {
     pub(crate) fn compile<'a>(
         schema: &'a Value,
         skip_prefix: usize,
-        context: &CompilationContext,
+        ctx: &compiler::Context,
     ) -> CompilationResult<'a> {
-        let keyword_context = context.with_path("items");
-        let node = compile_validators(schema, &keyword_context)?;
+        let ctx = ctx.new_at_location("items");
+        let node = compiler::compile(&ctx, ctx.as_resource_ref(schema))?;
         Ok(Box::new(ItemsObjectSkipPrefixValidator {
             node,
             skip_prefix,
@@ -165,6 +163,24 @@ impl ItemsObjectSkipPrefixValidator {
 }
 
 impl Validate for ItemsObjectSkipPrefixValidator {
+    #[allow(clippy::needless_collect)]
+    fn iter_errors<'i>(&self, instance: &'i Value, location: &LazyLocation) -> ErrorIterator<'i> {
+        if let Value::Array(items) = instance {
+            let errors: Vec<_> = items
+                .iter()
+                .skip(self.skip_prefix)
+                .enumerate()
+                .flat_map(move |(idx, item)| {
+                    self.node
+                        .iter_errors(item, &location.push(idx + self.skip_prefix))
+                })
+                .collect();
+            Box::new(errors.into_iter())
+        } else {
+            no_error()
+        }
+    }
+
     fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Array(items) = instance {
             items
@@ -176,37 +192,25 @@ impl Validate for ItemsObjectSkipPrefixValidator {
         }
     }
 
-    #[allow(clippy::needless_collect)]
-    fn validate<'instance>(
+    fn validate<'i>(
         &self,
-        instance: &'instance Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'instance> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
         if let Value::Array(items) = instance {
-            let errors: Vec<_> = items
-                .iter()
-                .skip(self.skip_prefix)
-                .enumerate()
-                .flat_map(move |(idx, item)| {
-                    self.node
-                        .validate(item, &instance_path.push(idx + self.skip_prefix))
-                })
-                .collect();
-            Box::new(errors.into_iter())
-        } else {
-            no_error()
+            for (idx, item) in items.iter().skip(self.skip_prefix).enumerate() {
+                let item_location = location.push(idx + self.skip_prefix);
+                self.node.validate(item, &item_location)?;
+            }
         }
+        Ok(())
     }
 
-    fn apply<'a>(
-        &'a self,
-        instance: &Value,
-        instance_path: &InstancePath,
-    ) -> PartialApplication<'a> {
+    fn apply(&self, instance: &Value, location: &LazyLocation) -> PartialApplication {
         if let Value::Array(items) = instance {
             let mut results = Vec::with_capacity(items.len().saturating_sub(self.skip_prefix));
-            for (idx, item) in items.iter().skip(self.skip_prefix).enumerate() {
-                let path = instance_path.push(idx + self.skip_prefix);
+            for (idx, item) in items.iter().enumerate().skip(self.skip_prefix) {
+                let path = location.push(idx);
                 results.push(self.node.apply_rooted(item, &path));
             }
             let mut output: PartialApplication = results.into_iter().collect();
@@ -214,7 +218,7 @@ impl Validate for ItemsObjectSkipPrefixValidator {
             // we must produce an annotation with a boolean value indicating whether the subschema
             // was applied to any positions in the underlying array.
             let schema_was_applied = items.len() > self.skip_prefix;
-            output.annotate(serde_json::json! {schema_was_applied}.into());
+            output.annotate(serde_json::json!(schema_was_applied).into());
             output
         } else {
             PartialApplication::valid_empty()
@@ -222,30 +226,23 @@ impl Validate for ItemsObjectSkipPrefixValidator {
     }
 }
 
-impl core::fmt::Display for ItemsObjectSkipPrefixValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "items: {}", format_validators(self.node.validators()))
-    }
-}
-
 #[inline]
 pub(crate) fn compile<'a>(
+    ctx: &compiler::Context,
     parent: &'a Map<String, Value>,
     schema: &'a Value,
-    context: &CompilationContext,
 ) -> Option<CompilationResult<'a>> {
     match schema {
-        Value::Array(items) => Some(ItemsArrayValidator::compile(items, context)),
+        Value::Array(items) => Some(ItemsArrayValidator::compile(ctx, items)),
         Value::Object(_) | Value::Bool(false) => {
-            #[cfg(feature = "draft202012")]
             if let Some(Value::Array(prefix_items)) = parent.get("prefixItems") {
                 return Some(ItemsObjectSkipPrefixValidator::compile(
                     schema,
                     prefix_items.len(),
-                    context,
+                    ctx,
                 ));
             }
-            Some(ItemsObjectValidator::compile(schema, context))
+            Some(ItemsObjectValidator::compile(ctx, schema))
         }
         _ => None,
     }
@@ -259,8 +256,8 @@ mod tests {
 
     #[test_case(&json!({"items": false}), &json!([1]), "/items")]
     #[test_case(&json!({"items": {"type": "string"}}), &json!([1]), "/items/type")]
-    #[test_case(&json!({"items": [{"type": "string"}]}), &json!([1]), "/items/0/type")]
-    fn schema_path(schema: &Value, instance: &Value, expected: &str) {
-        tests_util::assert_schema_path(schema, instance, expected)
+    #[test_case(&json!({"prefixItems": [{"type": "string"}]}), &json!([1]), "/prefixItems/0/type")]
+    fn location(schema: &Value, instance: &Value, expected: &str) {
+        tests_util::assert_schema_location(schema, instance, expected);
     }
 }
