@@ -1,12 +1,12 @@
 use crate::{
-    compilation::{compile_validators, context::CompilationContext},
+    compiler,
     error::{no_error, ErrorIterator, ValidationError},
     keywords::CompilationResult,
+    node::SchemaNode,
     output::BasicOutput,
-    paths::{InstancePath, JSONPointer},
-    primitive_type::PrimitiveType,
-    schema_node::SchemaNode,
-    validator::{format_key_value_validators, PartialApplication, Validate},
+    paths::{LazyLocation, Location},
+    types::JsonType,
+    validator::{PartialApplication, Validate},
 };
 use serde_json::{Map, Value};
 
@@ -16,34 +16,51 @@ pub(crate) struct PropertiesValidator {
 
 impl PropertiesValidator {
     #[inline]
-    pub(crate) fn compile<'a>(
-        schema: &'a Value,
-        context: &CompilationContext,
-    ) -> CompilationResult<'a> {
+    pub(crate) fn compile<'a>(ctx: &compiler::Context, schema: &'a Value) -> CompilationResult<'a> {
         match schema {
             Value::Object(map) => {
-                let context = context.with_path("properties");
+                let ctx = ctx.new_at_location("properties");
                 let mut properties = Vec::with_capacity(map.len());
                 for (key, subschema) in map {
-                    let property_context = context.with_path(key.clone());
+                    let ctx = ctx.new_at_location(key.as_str());
                     properties.push((
                         key.clone(),
-                        compile_validators(subschema, &property_context)?,
+                        compiler::compile(&ctx, ctx.as_resource_ref(subschema))?,
                     ));
                 }
                 Ok(Box::new(PropertiesValidator { properties }))
             }
             _ => Err(ValidationError::single_type_error(
-                JSONPointer::default(),
-                context.clone().into_pointer(),
+                Location::new(),
+                ctx.location().clone(),
                 schema,
-                PrimitiveType::Object,
+                JsonType::Object,
             )),
         }
     }
 }
 
 impl Validate for PropertiesValidator {
+    #[allow(clippy::needless_collect)]
+    fn iter_errors<'i>(&self, instance: &'i Value, location: &LazyLocation) -> ErrorIterator<'i> {
+        if let Value::Object(item) = instance {
+            let errors: Vec<_> = self
+                .properties
+                .iter()
+                .flat_map(move |(name, node)| {
+                    let option = item.get(name);
+                    option.into_iter().flat_map(move |item| {
+                        let instance_path = location.push(name.as_str());
+                        node.iter_errors(item, &instance_path)
+                    })
+                })
+                .collect();
+            Box::new(errors.into_iter())
+        } else {
+            no_error()
+        }
+    }
+
     fn is_valid(&self, instance: &Value) -> bool {
         if let Value::Object(item) = instance {
             self.properties.iter().all(move |(name, node)| {
@@ -55,47 +72,34 @@ impl Validate for PropertiesValidator {
         }
     }
 
-    #[allow(clippy::needless_collect)]
-    fn validate<'instance>(
+    fn validate<'i>(
         &self,
-        instance: &'instance Value,
-        instance_path: &InstancePath,
-    ) -> ErrorIterator<'instance> {
+        instance: &'i Value,
+        location: &LazyLocation,
+    ) -> Result<(), ValidationError<'i>> {
         if let Value::Object(item) = instance {
-            let errors: Vec<_> = self
-                .properties
-                .iter()
-                .flat_map(move |(name, node)| {
-                    let option = item.get(name);
-                    option.into_iter().flat_map(move |item| {
-                        let instance_path = instance_path.push(name.clone());
-                        node.validate(item, &instance_path)
-                    })
-                })
-                .collect();
-            Box::new(errors.into_iter())
-        } else {
-            no_error()
+            for (name, node) in &self.properties {
+                if let Some(item) = item.get(name) {
+                    node.validate(item, &location.push(name))?;
+                }
+            }
         }
+        Ok(())
     }
 
-    fn apply<'a>(
-        &'a self,
-        instance: &Value,
-        instance_path: &InstancePath,
-    ) -> PartialApplication<'a> {
+    fn apply(&self, instance: &Value, location: &LazyLocation) -> PartialApplication {
         if let Value::Object(props) = instance {
             let mut result = BasicOutput::default();
             let mut matched_props = Vec::with_capacity(props.len());
             for (prop_name, node) in &self.properties {
                 if let Some(prop) = props.get(prop_name) {
-                    let path = instance_path.push(prop_name.clone());
+                    let path = location.push(prop_name.as_str());
                     matched_props.push(prop_name.clone());
                     result += node.apply_rooted(prop, &path);
                 }
             }
             let mut application: PartialApplication = result.into();
-            application.annotate(serde_json::Value::from(matched_props).into());
+            application.annotate(Value::from(matched_props).into());
             application
         } else {
             PartialApplication::valid_empty()
@@ -103,26 +107,16 @@ impl Validate for PropertiesValidator {
     }
 }
 
-impl core::fmt::Display for PropertiesValidator {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(
-            f,
-            "properties: {{{}}}",
-            format_key_value_validators(&self.properties)
-        )
-    }
-}
-
 #[inline]
 pub(crate) fn compile<'a>(
+    ctx: &compiler::Context,
     parent: &'a Map<String, Value>,
     schema: &'a Value,
-    context: &CompilationContext,
 ) -> Option<CompilationResult<'a>> {
     match parent.get("additionalProperties") {
         // This type of `additionalProperties` validator handles `properties` logic
-        Some(Value::Bool(false)) | Some(Value::Object(_)) => None,
-        _ => Some(PropertiesValidator::compile(schema, context)),
+        Some(Value::Bool(false) | Value::Object(_)) => None,
+        _ => Some(PropertiesValidator::compile(ctx, schema)),
     }
 }
 
@@ -132,11 +126,11 @@ mod tests {
     use serde_json::json;
 
     #[test]
-    fn schema_path() {
-        tests_util::assert_schema_path(
+    fn location() {
+        tests_util::assert_schema_location(
             &json!({"properties": {"foo": {"properties": {"bar": {"required": ["spam"]}}}}}),
             &json!({"foo": {"bar": {}}}),
             "/properties/foo/properties/bar/required",
-        )
+        );
     }
 }
